import logging
import operator

import numpy as np
import pandas as pd
import shap

__all__ = [
    "append_empty_shap_columns",
    "calculate_shap_percentile",
    "filter_rows_with_index",
    "generate_shap_top_factors",
    "reset_multiindex",
    "select_index",
    "shap_score_to_percentile",
]

logger = logging.getLogger(__file__)
MAX_FEATURES = 10


def get_shap_factor_values(row, shap_score_dict, key):
    if key == "pos":
        reverse = True
    else:
        reverse = False

    sorted_shap_score_dict = sorted(
        shap_score_dict.items(), key=operator.itemgetter(1), reverse=reverse
    )

    # loop thru top factors and extract factor name, its shap value, and the patient's
    # value for that factor
    num_features = min(len(sorted_shap_score_dict), MAX_FEATURES)

    factors = [sorted_shap_score_dict[idx][0][:-11] for idx in range(num_features)]
    shap_scores = [
        round(row[sorted_shap_score_dict[idx][0]], 3) for idx in range(num_features)
    ]
    patient_values = [
        row[sorted_shap_score_dict[idx][0][:-11]] for idx in range(num_features)
    ]

    row[key + "_factors"] = factors
    row[key + "_shap_scores"] = np.array(shap_scores)
    row[key + "_patient_values"] = patient_values

    return row


# function to construct top factor df in a vectorized way
def build_top_factor_df(row, shap_base_value):
    # build dictionary for positive and negative shap_scores
    pos_shap_score_dict = {}
    neg_shap_score_dict = {}

    for k, v in row.to_dict().items():
        if v is not None and k.endswith("_shap_score") and v > 0.0:
            pos_shap_score_dict[k] = v
        elif v is not None and k.endswith("_shap_score") and v < 0.0:
            neg_shap_score_dict[k] = v

    row = get_shap_factor_values(row, pos_shap_score_dict, "pos")
    row = get_shap_factor_values(row, neg_shap_score_dict, "neg")

    return row


def unmap_int_cols(df, outcome_column, mapping):
    """
    function to unmap categorical columns

    df - dataframe
    mapping - mapping dictionary

    returns: df with categorical columns mapped back to original values
    """
    for key in mapping.keys():

        # build the unmapping dictionary for "key"
        unmap_dict = {v: k for k, v in mapping[key].items()}

        # apply the unmapping to the items in column "key"
        if key != outcome_column:
            # df[key] = df[key].apply(lambda x: unmap_dict[x])
            df[key] = df[key].apply(
                lambda x: unmap_dict[x] if (np.all(pd.notnull(x))) else None
            )

    return df


def append_empty_shap_columns(df):
    factor_column_list = [
        "pos_factors",
        "pos_shap_scores",
        "pos_patient_values",
        "pos_shap_scores_w",
        "pos_shap_percentile" "neg_factors",
        "neg_shap_scores",
        "neg_patient_values",
        "neg_shap_scores_w",
        "neg_shap_percentile",
    ]
    # Make a column containing an empty list for every row.
    l = []
    v = df[df.columns[0]].apply(lambda x: l)
    for c in factor_column_list:
        df[c] = v
    return df


def generate_shap_top_factors(
    df, model, outcome_column, mapping, shap_approximate=False, **kwargs
):
    """Computes the contributing factors using the SHAP package and orders 
    them by absolute value

    Input: feature dataframe and the xgboost model object
    Output: dataframe of ordered contributing factors,
            their shap value and std devs from population mean
            and the patient's feature value
    """
    # compute the shap values
    logger.warning(f"Computing SHAP scores.  Approximate = {shap_approximate}")
    shap_values = shap.TreeExplainer(model).shap_values(
        df, approximate=shap_approximate, check_additivity=False
    )
    logger.warning(f"SHAP values completed")
    shap_df = pd.DataFrame(shap_values)
    #logger.info(f"SHAP: {shap_df.shape[0]} {df.shape[0]}")

    shap_base_value = shap_df.iloc[0, -1]

    # drop the bias term (last column) - unless it's not present.
    if shap_df.shape[1] == df.shape[1] + 1:
        shap_df.drop(shap_df.shape[1] - 1, axis=1, inplace=True)

    assert shap_df.shape[0] == df.shape[0], (
        f"shap_values was {shap_df.shape}, didn't have the same number of rows"
        f" as {df.shape}"
    )
    assert shap_df.shape[1] == df.shape[1], (
        f"shap_values was {shap_df.shape}, did not have the same number of columns"
        f" as {df.shape} {shap_df.columns.tolist()}"
    )

    # make a dict for feature name
    df_columns = df.columns.tolist()
    shap_df_columns = shap_df.columns.tolist()
    feature_map = {
        shap_df_columns[i]: df_columns[i] + "_shap_score"
        for i in range(len(df_columns))
    }

    # rename columns to align with passed in df
    shap_df.rename(columns=feature_map, inplace=True)
    shap_df.index = df.index

    # join original data with shap_df
    assert shap_df.shape[0] == df.shape[0]
    assert len(set(shap_df.index.values)) == df.shape[0]
    assert len(set(df.index.values)) == df.shape[0]
    assert (shap_df.index.values == df.index.values).all()
    #joined_df = pd.merge(df, shap_df, left_index=True, right_index=True)
    joined_df = df.join(shap_df)
    assert joined_df.shape[0] == df.shape[0], f"{joined_df.shape[0]} == {df.shape[0]}"

    # unmap categorical columns
    joined_df = unmap_int_cols(joined_df, outcome_column, mapping)

    # initialize top factor columns (capped at MAX_FEATURES)
    num_features = len(shap_df_columns)
    if num_features > MAX_FEATURES:
        num_features = MAX_FEATURES
    factor_column_list = [
        "pos_factors",
        "pos_shap_scores",
        "pos_patient_values",
        "neg_factors",
        "neg_shap_scores",
        "neg_patient_values",
    ]

    # use build_top_factor_df function to compute top factor info
    joined_df = joined_df.apply(build_top_factor_df, args=(shap_base_value,), axis=1)

    # keep only the top factor columns
    top_factor_df = joined_df[factor_column_list]

    return top_factor_df, shap_base_value


def calculate_shap_percentile(pred):
    """Computes the percentile of a distribution of SHAP scores
        input: dataframe with pos and neg shap score column_names
        output: array of percentiles for abs() value of all shap scores
    """

    pos_shap_scores = np.abs(np.hstack(pred["pos_shap_scores"]))
    neg_shap_scores = np.abs(np.hstack(pred["neg_shap_scores"]))
    all_shap_scores = np.concatenate([pos_shap_scores, neg_shap_scores])

    q = np.array(range(100))

    if len(all_shap_scores) == 0:
        all_shap_scores = [0.0]
    shap_pct = np.percentile(all_shap_scores, q)

    return shap_pct


def filter_rows_with_index(r, cutoff):
    """Return index of items in the list above the cutoff"""
    return [i for i, x in enumerate(r) if abs(x) > cutoff]


def select_index(r, i):
    """Return the items in the list from the specified index list"""
    return list(np.take(r, i))


def reset_multiindex(df):
    return df.reset_index(drop=True, level=0)



from cv19index.io import read_demographics, read_claim
from .predict import do_run_claims, do_run
from .preprocess import preprocess_xgboost
from datetime import datetime
import pandas as pd


def test_do_run_claims():
    do_run_claims('examples/demographics.csv',
                  'examples/claims.csv',
                  f'predictions-{datetime.now().strftime("%Y-%M-%dT%H-%m-%S")}.csv',
                  'xgboost_all_ages',
                  pd.to_datetime('2018-12-01'))

def test_do_run_claims_xlsx():
    do_run_claims('examples/demographics.xlsx',
                  'examples/claims.xlsx',
                  f'predictions-{datetime.now().strftime("%Y-%M-%dT%H-%m-%S")}.csv',
                  'xgboost_all_ages',
                  pd.to_datetime('2018-12-01'))



# IO functions to read and write model and prediction files in an appropriate format.

import json
import logging
import math
import pickle
from typing import Dict
import pandas as pd
from pandas.io.common import _NA_VALUES
from pkg_resources import resource_filename

from .util import schema_dtypes

logger = logging.getLogger(__name__)

INDEX = 'personId'


def read_model(fpath):
    with open(fpath, "rb") as fobj:
        return pickle.load(fobj)


def get_na_values(dtypes):
    # Pandas converts a string with "NA" as a real NaN/Null. We don't want this
    # for real string columns. NA can show up as a real flag in data and
    # it doesn't mean it should be treated as NaN/Null.
    def na_vals(x):
        # For now, only ignore NA conversion for strings. Structs/etc can still use it.
        if x in ("string", ):
            return []
        else:
            return _NA_VALUES

    return {k: na_vals(v) for k, v in dtypes.items()}


def read_claim(fpath: str) -> pd.DataFrame:
    schame_fpath = resource_filename("cv19index", "resources/claims.schema.json")
    return read_frame(fpath, schame_fpath, date_cols = ['admitDate', 'dischargeDate'])

import collections

def read_demographics(fpath: str) -> pd.DataFrame:
    schema_fpath = resource_filename("cv19index", "resources/demographics.schema.json")
    df = read_frame(fpath, schema_fpath).set_index('personId')
    duplicates = [x[0] for x in collections.Counter(df.index.values).items() if x[1] > 1]
    assert len(duplicates) == 0, f"Duplicate person ids in demographics file: {duplicates[:5]}"
    return df


def read_frame(fpath, schema_fpath, date_cols = []) -> pd.DataFrame:
    XLS = ('xls', 'xlsx', 'xlsm', 'xlsb', 'odf')

    def is_excel(fpath: str) -> bool:
        return fpath.endswith(XLS)

    with open(schema_fpath) as f:
        schema = json.load(f)
    dtypes = schema_dtypes(schema["schema"])

    if is_excel(fpath):
        df = read_excel(fpath, dtypes, date_cols)
    elif fpath.endswith(".parquet"):
        df = read_parquet(fpath, dtypes, date_cols)
    elif fpath.endswith(".csv"):
        df = read_csv(fpath, dtypes, date_cols)
    else:
        raise TypeError(f'This script reads files based the extension.\n'
                        f'The known extensions are {", ".join(XLS)}, .parquet, .csv'
                        f'Please ensure your file is one of those file types with correct file extension.')

    return df


def read_excel(fpath, dtype, date_cols) -> pd.DataFrame:
    na_values = get_na_values(dtype)

    df = pd.read_excel(
        fpath,
        header=0,
        dtype=dtype,
        parse_dates=date_cols,
        na_values=na_values,
        keep_default_na=False
    )
    for c in date_cols:
        df[c] = pd.to_datetime(df[c], infer_datetime_format = True)

    return df


def read_parquet(fpath, dtype, date_cols) -> pd.DataFrame:
    df = pd.read_parquet(fpath)
    # Now set the index.
    df = df.set_index(INDEX)
    return df


def read_csv(fpath: str, dtype: Dict, date_cols) -> pd.DataFrame:
    na_values = get_na_values(dtype)

    # Pandas won't read in ints with NA values, so read those in as floats.
    def adj_type(x):
        if x == "int32" or x == "int64":
            return "float64"
        return x

    df = pd.read_csv(
        fpath,
        header=0,
        na_values=na_values,
        index_col=False,
        keep_default_na=False,
    )

    # Oh, and if there are date/datetime types that are empty, pandas "helps us out"
    # by converting those to an int(0) instead of None.
    # Thanks pandas, but I'd prefer None here.
    for c in date_cols:
        df[c] = pd.to_datetime(df[c], infer_datetime_format = True)

    return df


def write_predictions(predictions, fpath):
    if fpath.endswith(".csv"):
        output = predictions.to_csv(index=False, float_format="%f")
    elif fpath.endswith(".json"):
        if (
            predictions.index.dtypes == "object"
            and predictions.index.size > 0
            and type(predictions.index[0]) == list
        ):
            predictions.index = [tuple(l) for l in predictions.index]
        js_rec = predictions.to_json(orient="records", double_precision=3)
        output = f'{{"records":{js_rec}}}'
    elif fpath.endswith(".jsonl"):
        # These next two lines are because there was a bug where to_json would
        # fail if called with an index that is an array, which happens with
        # compound ids.  Since writing "records" doesn't write the index, we
        # throw it away.
        if predictions.index.dtypes == "object":
            predictions = predictions.reset_index(drop=True)
        output = predictions.to_json(orient="records", lines=True, double_precision=3)
    else:
        raise Exception(
            f"Unsupported output format for {fpath}.  Must be .csv, .json, or .jsonl"
        )
    with open(fpath, "wt") as fobj:
        fobj.write(output)


def _eval_array_column(x):
    """Takes a column which is a JSON string and converts it into an array."""
    if type(x) == str:
        a = eval(x)
        if type(a) != list:
            raise Exception(f"Unexpected data in array column: {x}")
        return a
    elif x is None or (type(x) == float and math.isnan(x)):
        return []
    else:
        msg = f"Unexpected data in array column: {x}"
        raise Exception(msg)



import math

import numpy as np
import pandas as pd

_defaultDtypeDict = {
    "string": "object",
    "struct": "object",
    "boolean": "bool",
    "double": "float64",
    "integer": "int32",
    "long": "int64",
    "date": "object",
    "timestamp": "object",
    "datetime": "object",
    "array": "object",
}


class UserException(Exception):
    """Raise for exceptions that indicate bad input """


def schema_dtypes(schema_json: list, dtype_dict=None):
    if dtype_dict is None:
        dtype_dict = _defaultDtypeDict

    def lookup(x):
        t = dtype_dict.get(x["dataType"]["dataType"])
        assert len(t) > 0
        return t

    return {x["name"]: lookup(x) for x in schema_json}


def nonnull_column(s: pd.Series):
    """ Takes a pandas Series and return sa vector of all entries that have 
    values (are not None or float.nan).
    """
    if s.dtype == float:
        return np.logical_not(np.isnan(s))
    else:
        return np.logical_not(pd.isnull(s))


def none_or_nan(v):
    """Returns True if the value is None or float('nan')"""
    return v is None or (isinstance(v, float) and math.isnan(v))



import tempfile
import time
import pandas as pd
import numpy as np

from pkg_resources import resource_filename
from flask import Flask, make_response, request, abort

from .predict import do_run
from .io import read_model


def create_model_app(model_fpath, schema_fpath, **kwargs):
    model = read_model(model_fpath)

    app = Flask(__name__)

    @app.route("/ping", methods=("GET",))
    def ping():
        return make_response("")

    @app.route("/invocations", methods=("POST",))
    def invocations():
        start = time.time()

        with tempfile.NamedTemporaryFile(suffix=".csv") as input_fobj:
            input_fpath = input_fobj.name

            input_fobj.write(request.data)
            input_fobj.seek(0)

            with tempfile.NamedTemporaryFile(suffix=".csv") as output_fobj:
                output_fpath = output_fobj.name

                try:
                    do_run(input_fpath, schema_fpath, model_fpath, output_fpath)
                except Exception as e:
                    abort(400, f"Error parsing payload: {e}")

                output_fobj.seek(0)
                results = output_fobj.read()
                response = make_response(results)
                response.mimetyhpe = "text/csv"
                return response

    return app


def sagemaker_serve():
    app = create_model_app(
        resource_filename("cv19index", "resources/xgboost/model.pickle"),
        resource_filename("cv19index", "resources/xgboost/input.csv.schema.json"),
    )
    app.run("0.0.0.0", 8080, debug=False)



import json
import logging
import math
import regex as re
from datetime import datetime

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from .util import none_or_nan, nonnull_column

logger = logging.getLogger(__name__)


def apply_int_mapping(mapping, data, error_unknown_values=True):
    """Maps strings to integers in a Data Frame using a prespecified mapping.
    Takes a mapping of values and applies it to the data.
    Null or None values will be mapped over.
    Params:
        mapping - A nested dictionary of {colName: {'value':mapped_value}}
        data - The DataFrame to map
    Returns:
        A new DataFrame with the string columns mapped.
    """
    ret = data.loc[:, list(set(data.columns) - set(mapping.keys()))].copy()
    for col, value_map in mapping.items():
        if col not in data.columns:
            if error_unknown_values:
                logger.error(f'No column named "{col}" in {data.columns}')
                raise ValueError(f'No column named "{col}" in {data.columns}')
            else:
                continue

        ret[col] = data[col].map(value_map)

        # Check if any non-null values in the original array are not mapped
        if error_unknown_values and np.any(
                np.logical_and(np.isnan(ret[col]), nonnull_column(data[col]))
        ):
            raise ValueError(
                f"Column '{col}' had invalid values:"
                f" {set(data[col].unique()) - set(value_map.keys())}."
                f"  Valid values are {set(value_map.keys())}"
            )
    return ret


def cleanICD10Syntax(code):
    code = str(code)
    if len(code) > 3 and '.' not in code:
        return code[:3] + '.' + code[3:]
    else:
        return code


def preprocess_xgboost(claim_df: pd.DataFrame, demo_df: pd.DataFrame, asOfDate: pd.datetime):
    DIAGNOSIS_COLS = ['dx1', 'dx2', 'dx3', 'dx4', 'dx5', 'dx6', 'dx7', 'dx8', 'dx9', 'dx10', 'dx11', 'dx12', 'dx13',
                      'dx14', 'dx15', 'dx16']
    preprocessed_df = demo_df.loc[:,['gender', 'age']].rename(columns={'gender': 'Gender', 'age': 'Age'})

    logger.debug(f"Beginning claims data frame preprocessing, raw data frame as follows.")
    logger.debug(claim_df.head(5))

    asOfPastYear = str(pd.to_datetime(asOfDate) - pd.DateOffset(years=1))[:10]

    # limit to last year of claims
    orig_len = claim_df.shape[0]
    claim_df = claim_df[(asOfPastYear <= claim_df['admitDate'])
                        & (claim_df['admitDate'] <= asOfDate)]
    logger.debug(f"Filtered claims to just those within the dates {asOfPastYear} to {asOfDate}.  Claim count went from {orig_len} to {len(claim_df)}")

    # total numbers of days in the ER
    er_visit = claim_df.loc[claim_df['erVisit'] == True, :].groupby('personId').admitDate.nunique()
    preprocessed_df['# of ER Visits (12M)'] = er_visit

    inpatient_rows = claim_df.loc[claim_df['inpatient'] == True, :]
    preprocessed_df['# of Admissions (12M)'] = inpatient_rows.groupby('personId').admitDate.nunique()
    inpatient_days = pd.Series((inpatient_rows['dischargeDate'].dt.date - inpatient_rows['admitDate'].dt.date).dt.days, index=claim_df['personId'])
    preprocessed_df['Inpatient Days'] = inpatient_days.groupby('personId').sum()

    # Cleaning the diagnosis codes to apply to all the dx cols
    logger.debug(f"Cleaning diagnosis codes.")
    used_diags = [x for x in DIAGNOSIS_COLS if x in claim_df.columns]
    for column in used_diags:
        claim_df[column] = claim_df.loc[:,column].map(cleanICD10Syntax, na_action='ignore')

    # Generating features for each node
    logger.debug(f"Computing diagnosis flags.")
    nodes = pd.read_csv(resource_filename('cv19index', 'resources/ccsrNodes.txt'))
    edges_df = pd.read_csv(resource_filename('cv19index', 'resources/ccsrEdges.txt'))
    edges_df['code'] = edges_df['child'].apply(lambda x: x.split(':')[1])

    for CCSR, description in nodes.values:
        # Getting the codes
        codes = set(edges_df[edges_df['parent'].str.contains(CCSR)]['code'].values)
        #logger.debug(f"Codes are {codes}")
        matches = claim_df.loc[:,used_diags].isin(codes).any(axis=1)
        #logger.debug(f"Matches are {matches.shape[0]}\n{matches.head()}")
        selected_claim = claim_df.loc[matches, 'personId']
        #logger.debug(f"selected_claim are {selected_claim.shape[0]}\n{selected_claim.head()}")
        selected_personId = np.unique(selected_claim)
        #logger.debug(f"Selected are {selected_personId.shape[0]}")

        # Assigning the diagnosis flag to the person
        description = re.sub("[^\P{P}-/']+", "_", description.replace(")", ""))
        column_name = "Diagnosis of " + description + " in the previous 12 months"
        preprocessed_df[column_name] = pd.Series(True, index=selected_personId, dtype=np.bool)
        preprocessed_df[column_name] = preprocessed_df[column_name].fillna(False)
        #preprocessed_df[column_name] = preprocessed_df[column_name].astype(np.bool)
        #logger.debug(f"Final is\n{preprocessed_df[column_name]}")

    # fill in 0's
    preprocessed_df.fillna(0, inplace=True)

    logger.debug(f"Preprocessing complete.")
    return preprocessed_df



import argparse
import json
import logging
import math
import urllib
from typing import Any, Dict, List, Tuple
from datetime import datetime

import numpy as np
import pandas as pd
import xgboost as xgb
from pkg_resources import resource_filename

from .io import read_frame, read_model, write_predictions, read_claim, read_demographics
from .preprocess import apply_int_mapping, preprocess_xgboost
from .shap_top_factors import (
    append_empty_shap_columns,
    calculate_shap_percentile,
    filter_rows_with_index,
    generate_shap_top_factors,
    reset_multiindex,
    select_index
)

logger = logging.getLogger(__file__)

PREDICTION_QUANTILES = "prediction_quantiles"


def clean_floats(d: pd.DataFrame):
    new_d = d.copy()
    for k, v in new_d.items():
        if math.isnan(v) or math.isinf(v):
            new_d[k] = None
        elif isinstance(v, np.float64):
            new_d[k] = float(v)
    return new_d


def rescale_predictions(
        predictions: np.ndarray, train_data_stats: Dict[str, Any]
) -> np.ndarray:
    """
    Function to rescale model predictions s.t. they correspond to probabilities.
    This funciton is only called if pos_scale_weight > 1.  Some background can be
    found here: https://github.com/dmlc/xgboost/issues/863

    predictions - the original predictions which need to be rescaled
    train_data_stats - dictionary containing the total number of events and number
        of rare events

    Returns: rescaled predictions which now correspond to probabilities
    """

    # original total number of events
    E = train_data_stats["total_events"]
    # original number of rare events
    B = train_data_stats["rare_events"]
    # original number of non-rare events
    A = E - B
    # new number of non-rare events
    C = A
    # new number of rare events
    D = A
    # new total number of events
    F = C + D

    fac1 = (A / E) / (C / F)
    fac2 = (B / E) / (D / F)

    def rescale_predictions(p):
        a0 = (1 - p) * fac1

        a1 = p * fac2

        return a1 / (a1 + a0)

    predictions = np.apply_along_axis(rescale_predictions, 0, predictions)

    return predictions


def perform_predictions(
        df: pd.DataFrame,
        xmatrix: xgb.DMatrix,
        label: np.ndarray,
        predictor: Dict[str, Any],
        recompute_distribution: bool,
        shap_score_99: float = None,
        shap_cutoff: float = 0.02,
        compute_factors: bool = True,
        factor_cutoff: float = None,
        **kwargs,
) -> Tuple[pd.DataFrame, List[float], float, float]:
    """
    Build predictions from a trained model, additionally build shap scores and quantiles

    Args:
        compute_factors_at_cutoff: float (0, 1], only compute SHAP for rows
        beyond this cutoff after sorted by prediction

    return prediction DataFrame, quantiles, shap_base_value, shap_score_99
    """
    outcome_column = predictor["outcome_column"]
    mapping = predictor["mapping"]
    model = predictor["model"]

    predictions = model.predict(xmatrix)
    assert len(predictions) == df.shape[0]

    # rescale the predictions if model used scale_pos_weight_flag
    if predictor["predictor_type"] == "classification":
        if predictor["hyper_params"]["scale_pos_weight"] > 1:
            logger.info(
                f'Scale pos weight is {predictor["hyper_params"]["scale_pos_weight"]}.'
                " Rescaling predictions to probabilities"
            )
            predictions = rescale_predictions(
                predictions, predictor["train_data_stats"]
            )
        else:
            logger.info(
                f'Scale pos weight is {predictor["hyper_params"]["scale_pos_weight"]}.'
                " No need to rescale predictions"
            )

    # If our index is a list type, convert it to tuples
    index_name = df.index.name
    if df.index.dtype == "object" and df.index.size > 0 and type(df.index[0]) == list:
        df.index = [tuple(l) for l in df.index]
        df.index.name = index_name

    prediction_quantiles = get_quantiles(predictions, predictor, recompute_distribution)
    risk_scores = np.digitize(predictions, prediction_quantiles)

    # need to include the label in the output df for split trains
    # in the case of a composite key DataFrame joining doesnt naively work
    # instead including it at construction time
    pred_dict = {
        df.index.name: df.index,
        "prediction": predictions,
        "risk_score": risk_scores,
    }
    if label is not None:
        pred_dict[outcome_column] = label
    prediction_result = pd.DataFrame(pred_dict, index=df.index)

    assert prediction_result.shape[0] == df.shape[0]

    # compute top factors and "base value" using SHAP technique
    top_factors, shap_base_value = (None, None)
    if compute_factors:
        if isinstance(factor_cutoff, float):
            compute_factors_cutoff_val = prediction_quantiles[
                int(100 * (1 - factor_cutoff))
            ]
            df_cutoff = df[predictions >= compute_factors_cutoff_val]
            #logger.info(f"Cutoff df: {df_cutoff}")
            top_factors, shap_base_value = generate_shap_top_factors(
                df_cutoff, model, outcome_column, mapping, **kwargs
            )
            assert top_factors.shape[0] == df.shape[0], f"Found: {top_factors.shape[0]} Expected: {df.shape[0]}"
            logger.info(
                f"Computed SHAP scores for top {100 * factor_cutoff}% of predictions"
                f" resulting in {top_factors.shape[0]} scores."
            )
            assert prediction_result.shape[0] == df.shape[0]
            prediction_result = prediction_result.join(top_factors)
            assert prediction_result.shape[0] == df.shape[0], f"Found: {prediction_result.shape[0]} Expected: {df.shape[0]}"

            empty_list = np.empty(shape=(0,))
            for col_name in top_factors.columns:
                prediction_result[col_name] = prediction_result[col_name].apply(
                    lambda d: d
                    if (isinstance(d, list) or isinstance(d, np.ndarray))
                    else empty_list
                )
        else:
            top_factors, shap_base_value = generate_shap_top_factors(
                df, model, outcome_column, mapping, **kwargs
            )
            prediction_result = prediction_result.join(top_factors)

    # If index is tuples, convert it back to list
    if type(prediction_result.index[0]) == tuple:
        prediction_result[index_name] = [list(l) for l in prediction_result.index]
        prediction_result = prediction_result.set_index(index_name, drop=False)

    if compute_factors:
        # Calculate 99th percentile shap score if it doesn't already exist
        if shap_score_99 is None:
            abs_shap_scores = np.abs(
                np.concatenate(
                    prediction_result[
                        ["pos_shap_scores", "neg_shap_scores"]
                    ].values.flatten()
                )
            )
            if len(abs_shap_scores) < 1:
                shap_score_99 = 0.0
            else:
                shap_score_99 = np.percentile(abs_shap_scores, 99)

        logger.info(f"Cutoff for SHAP values: {shap_cutoff}")
        assert prediction_result.shape[0] == df.shape[0]

        prediction_result["neg_index_filter"] = prediction_result[
            "neg_shap_scores"
        ].apply(filter_rows_with_index, args=(shap_cutoff,))
        prediction_result["pos_index_filter"] = prediction_result[
            "pos_shap_scores"
        ].apply(filter_rows_with_index, args=(shap_cutoff,))

        prediction_result["neg_shap_scores_w"] = (
                prediction_result["neg_shap_scores"] / shap_score_99
        )
        prediction_result["pos_shap_scores_w"] = (
                prediction_result["pos_shap_scores"] / shap_score_99
        )

        prediction_result["neg_shap_scores"] = (
            reset_multiindex(prediction_result)[["neg_shap_scores", "neg_index_filter"]]
                .apply(lambda x: select_index(*x), axis=1)
                .values
        )
        prediction_result["neg_shap_scores_w"] = (
            reset_multiindex(prediction_result)[
                ["neg_shap_scores_w", "neg_index_filter"]
            ]
                .apply(lambda x: select_index(*x), axis=1)
                .values
        )
        prediction_result["neg_factors"] = (
            reset_multiindex(prediction_result)[["neg_factors", "neg_index_filter"]]
                .apply(lambda x: select_index(*x), axis=1)
                .apply(lambda x: [urllib.parse.unquote(col) for col in x])
                .values
        )
        prediction_result["neg_patient_values"] = (
            reset_multiindex(prediction_result)[
                ["neg_patient_values", "neg_index_filter"]
            ]
                .apply(lambda x: select_index(*x), axis=1)
                .values
        )

        prediction_result["pos_shap_scores"] = (
            reset_multiindex(prediction_result)[["pos_shap_scores", "pos_index_filter"]]
                .apply(lambda x: select_index(*x), axis=1)
                .values
        )
        prediction_result["pos_shap_scores_w"] = (
            reset_multiindex(prediction_result)[
                ["pos_shap_scores_w", "pos_index_filter"]
            ]
                .apply(lambda x: select_index(*x), axis=1)
                .values
        )
        prediction_result["pos_factors"] = (
            reset_multiindex(prediction_result)[["pos_factors", "pos_index_filter"]]
                .apply(lambda x: select_index(*x), axis=1)
                .apply(lambda x: [urllib.parse.unquote(col) for col in x])
                .values
        )
        prediction_result["pos_patient_values"] = (
            reset_multiindex(prediction_result)[
                ["pos_patient_values", "pos_index_filter"]
            ]
                .apply(lambda x: select_index(*x), axis=1)
                .values
        )

        prediction_result = prediction_result.drop(
            columns=["neg_index_filter", "pos_index_filter"]
        )
        assert prediction_result.shape[0] == df.shape[0]
    else:
        prediction_result = append_empty_shap_columns(prediction_result)

    return prediction_result, prediction_quantiles, shap_base_value, shap_score_99


def flatten_predictions(prediction: pd.DataFrame):
    def extract_col(name, ix):
        prediction[name].apply(lambda x: x[ix])

    pos_cols_to_flatten = ['pos_factors', 'pos_patient_values', 'pos_shap_scores']
    neg_cols_to_flatten = ['neg_factors', 'neg_patient_values', 'neg_shap_scores']

    for i in range(10):
        for name in pos_cols_to_flatten:
            prediction[f"{name}_{i+1}"] = prediction[name].apply(lambda x: x[i] if i < len(x) else None)
    for i in range(10):
        for name in neg_cols_to_flatten:
            prediction[f"{name}_{i+1}"] = prediction[name].apply(lambda x: x[i] if i < len(x) else None)

    cols_to_drop = ['pos_shap_scores_w', 'neg_shap_scores_w']
    prediction = prediction.drop(cols_to_drop + pos_cols_to_flatten + neg_cols_to_flatten, axis=1)
    return prediction

def get_quantiles(
        predictions: np.ndarray, predictor: Dict[str, Any], recompute_distribution: bool
) -> List[float]:
    """
    Helper to get the quantiles of a prediction DataFrame
    """
    if recompute_distribution:
        q = np.array(range(100))
        prediction_quantiles = list(np.percentile(predictions, q=q))
    else:
        prediction_quantiles = predictor[PREDICTION_QUANTILES]
    return sorted(prediction_quantiles)


def get_agg_preds(val_to_preds: Dict[str, Any]) -> pd.DataFrame:
    """
    Take the dictionary that maps the split key to the model output
    aggregate the prediction DataFrames together as well as join the label data
    which can then be used to build test results

    """
    agg_preds = None
    for val, results in val_to_preds.items():
        if agg_preds is None:
            agg_preds = results["predictions"]
        else:
            agg_preds = pd.concat((agg_preds, results["predictions"]))

    return agg_preds


def reorder_inputs(df_inputs: pd.DataFrame, predictor: Dict[str, Any]) -> pd.DataFrame:
    """This code reorders in the input data frame columns to match the expected order
    from training. If the lists of columns don't match, no error is raised, as this
    will cause issues later."""
    if set(predictor["model"].feature_names) == set(df_inputs.columns) and predictor[
        "model"
    ].feature_names != list(df_inputs.columns):
        logger.info("Reordering test inputs to match training.")
        return df_inputs.loc[:, predictor["model"].feature_names]
    return df_inputs


def run_xgb_model(run_df: pd.DataFrame, predictor: Dict, **kwargs) -> pd.DataFrame:
    import xgboost as xgb

    df_inputs = apply_int_mapping(
        predictor["mapping"], run_df, error_unknown_values=False
    )

    df_inputs.columns = [urllib.parse.quote(col) for col in df_inputs.columns]
    df_inputs = reorder_inputs(df_inputs, predictor)
    run = xgb.DMatrix(df_inputs)
    factor_cutoff = (
        kwargs["predict_factor_cutoff"] if "predict_factor_cutoff" in kwargs else 1.0
    )
    assert run.num_row() == run_df.shape[0]
    assert df_inputs.shape[0] == run_df.shape[0]

    predictions, prediction_quantiles, shap_base_value, _ = perform_predictions(
        df_inputs,
        run,
        None,
        predictor,
        recompute_distribution=False,
        shap_score_99=predictor["shap_score_99"],
        factor_cutoff=factor_cutoff,
        **kwargs,
    )
    assert predictions.shape[0] == run_df.shape[0]

    shap_pct = predictor.get("shap_pct")
    if shap_pct is None:
        shap_pct = calculate_shap_percentile(predictions)

    return predictions


def write_xgb_predictions(predictions: pd.DataFrame, summary_file):
    predictions = flatten_predictions(predictions)
    output = predictions.to_csv(index=False, float_format="%f")

    with open(summary_file, "wt") as fobj:
        fobj.write(output)


def do_run(
        input_fpath: str, schema_fpath: str, model_fpath: str, output_fpath: str, **kwargs
) -> None:
    """
    Deprecated run function.  Use do_run_claims instead.
    """
    logger.info(f"Running {model_fpath} using {input_fpath} to {output_fpath}")
    run_df = read_frame(input_fpath, schema_fpath)

    if not run_df.empty:
        model = read_model(model_fpath)
        out = run_xgb_model(run_df, model, **kwargs)
        write_predictions(out, output_fpath)
    else:
        # If there are no inputs to run predictions on, bypass
        # executing the model and output an successful empty file.
        write_predictions(run_df, output_fpath)


def do_run_claims(fdemo, fclaim, output, model_name, asOfDate, feature_file = None):
    """
    Reads in claims and demographics files along with a trained model and generates an output file with predictions.
    """
    demo_df = read_demographics(fdemo)
    claim_df = read_claim(fclaim)
    model_file = resource_filename('cv19index', f'resources/{model_name}/model.pickle')

    logger.info(f"Reading model from {model_file}.  Writing results to {output}")

    input_df = preprocess_xgboost(claim_df, demo_df, asOfDate)
    assert input_df.shape[0] == demo_df.shape[0], f"Demographics file had {demo_df.shape[0]} lines."
    if feature_file is not None:
        input_df.to_csv(feature_file)

    model = read_model(model_file)
    predictions = run_xgb_model(input_df, model)
    assert predictions.shape[0] == demo_df.shape[0], f"Predictions file didn't match demographics {predictions.shape[0]} != {demo_df.shape[0]} "

    write_xgb_predictions(predictions, output)
    logger.info(f"Done.  Wrote {predictions.shape[0]} predictions.")


def parser():

    parser = argparse.ArgumentParser(description="Runs the CV19 Vulnerability Index form the command line.  Takes two files as input and produces an output file containing the predictions.")
    parser.add_argument("demographics_file", help='Path to the file containing demographics')
    parser.add_argument("claims_file", help='Path to the file containing claims data')
    parser.add_argument("output_file", default=f'predictions-{datetime.now().strftime("%Y-%M-%dT%H:%m:%S")}.csv', help='Path to the output file containing the results of the predictions.')
    parser.add_argument("-f", "--feature-file", help='If specified, writes a CSV file containing the preprocessed features that will be fed to the model.')
    parser.add_argument("-m", "--model", choices=["xgboost", "xgboost_all_ages"], default="xgboost_all_ages", help="Name of the model to use.  Corresponds to a directory under cv19index/resources.")
    parser.add_argument("-a", "--as-of-date", default=pd.to_datetime(datetime.today().isoformat()), help="Claims data uses data from 1 year prior to the prediction date.  This defaults to the current date, but can be overridden with this argument.")

    return parser.parse_args()


def main():
    logging.basicConfig(level=logging.DEBUG)
    args = parser()
    do_run_claims(args.demographics_file, args.claims_file, args.output_file, args.model, args.as_of_date, args.feature_file)



import pathlib
from setuptools import find_packages
from distutils.core import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text("utf-8")

setup(
    name="cv19index",
    version="1.1.1",
    description="COVID-19 Vulnerability Index",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://cv19index.com",
    author="Dave Decaprio",
    author_email="dave.decaprio@closedloop.ai",
    license="BSD",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Framework :: Jupyter",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    package_data={
        "cv19index": [
            "resources/xgboost/input.csv.schema.json",
            "resources/xgboost/model.pickle",
            "resources/xgboost_all_ages/input.csv.schema.json",
            "resources/xgboost_all_ages/model.pickle",
            "resources/demo.schema.json",
            "resources/claims.schema.json",
            "resources/ccsrEdges.txt",
            "resources/ccsrNodes.txt",
        ]
    },
    entry_points={
        "console_scripts": [
            "cv19index=cv19index.predict:main",
            # serve is required to be exposed by the sagemaker API.
            "serve=cv19index.server:sagemaker_serve",
        ]
    },
    install_requires=[
        "numpy>=1.17.4",
        "pandas>=0.23.4",
        "setuptools>=40.2.0",
        "shap>=0.33",
        "xgboost>=1.0.1",
        "flask>=1.1.1",
        "regex>=2020.2.20",
        "xlrd >= 0.9.0"
    ],
)
